UNIT CalcFuncs;
{$ifdef FPC}{$mode objfpc}{$h+}{$INTERFACES CORBA}{$endif}
{$LONGSTRINGS OFF}


INTERFACE

USES
 Classes, SysUtils, Math, mseguiglob, msekeyboard;

TYPE
 Operation = (_none, _done, _add, _sub, _mul, _div);

 NoCalculator = CLASS
  PROTECTED
    internalText: string;

  PRIVATE
    FUNCTION  getText:  string;                   VIRTUAL;
    PROCEDURE setText   (CONST NewValue: string); VIRTUAL;

  PUBLIC
    PROCEDURE Clear;                          VIRTUAL;
    PROCEDURE Delete;                         VIRTUAL; ABSTRACT;
    PROCEDURE ChangeSign;                     VIRTUAL; ABSTRACT;
    PROCEDURE nextDigit (Digit: char);        VIRTUAL; ABSTRACT;
    PROCEDURE nextDigit (Digit: integer);     VIRTUAL; ABSTRACT; OVERLOAD;
    PROCEDURE Operate   (OpSign: Operation);  VIRTUAL; ABSTRACT;
    PROCEDURE setPoint;                       VIRTUAL; // Dummy only!
    PROCEDURE Process   (Key: char);          VIRTUAL;
    PROCEDURE Process   (CONST Expr: string); VIRTUAL;
    PROCEDURE Process   (Key: keyty);         VIRTUAL;
    PROCEDURE KeyAction (VAR KeyInfo: keyeventinfoty); VIRTUAL;

    PROPERTY Entry: string READ getText WRITE setText;
  END;

//////////////////////////

 IntCalculator = CLASS (NoCalculator)
  PROTECTED
    currOp:    Operation;
    currValue: integer;
    prevOp:    ARRAY [0..1] OF Operation;
    Stack:     ARRAY [0..1] OF integer;

  PRIVATE
    FUNCTION  getValue: integer;             VIRTUAL;
    PROCEDURE setValue  (NewValue: integer); VIRTUAL;

  PUBLIC
    PROCEDURE Clear;                         OVERRIDE;
    PROCEDURE Delete;                        OVERRIDE;
    PROCEDURE ChangeSign;                    OVERRIDE;
    PROCEDURE nextDigit (inChar: char);      OVERRIDE;
    PROCEDURE nextDigit (Digit: integer);    OVERRIDE;
    PROCEDURE Operate   (OpSign: Operation); OVERRIDE;

    PROPERTY Value: integer READ currValue WRITE setValue;
  END;

//////////////////////////

 FloatCalculator = CLASS (NoCalculator)
  PROTECTED
    currOp:    Operation;
    currValue: double;
    pointpos:  smallint;
    prevOp:    ARRAY [0..1] OF Operation;
    Stack:     ARRAY [0..1] OF double;
    decSign:   ARRAY [0..1] OF smallint;

  PRIVATE
    FUNCTION  getValue: double;              VIRTUAL;
    PROCEDURE setValue  (NewValue: double);  VIRTUAL;

  PUBLIC
    CONSTRUCTOR Create;

    PROCEDURE Clear;                         OVERRIDE;
    PROCEDURE Delete;                        OVERRIDE;
    PROCEDURE ChangeSign;                    OVERRIDE;
    PROCEDURE nextDigit (inChar: char);      OVERRIDE;
    PROCEDURE nextDigit (Digit: integer);    OVERRIDE;
    PROCEDURE Operate   (OpSign: Operation); OVERRIDE;
    PROCEDURE setPoint;                      OVERRIDE;

    PROPERTY Value:  double   READ currValue WRITE setValue;
    PROPERTY Places: smallint READ pointpos;
  END;

///////////////////////////////////////////////////////

IMPLEMENTATION

//USES

///////////////////////////////////////////////////////

FUNCTION NoCalculator.getText: string;
 BEGIN
   getText:= internalText;
 END;

PROCEDURE NoCalculator.setText (CONST NewValue: string);
 BEGIN
   internalText:= NewValue;
 END;

PROCEDURE NoCalculator.Clear;
 BEGIN
   internalText:= '';
 END;

PROCEDURE NoCalculator.setPoint;
 BEGIN
   // Dummy only!
 END;

PROCEDURE NoCalculator.Process (Key: char);
 BEGIN
   CASE Key OF
     '0'..'9': NextDigit (Key);
     '/':      Operate (_div);
     '*':      Operate (_mul);
     '-':      Operate (_sub);
     '+':      Operate (_add);
     #$7F:     Clear;       // Delete
     ^H:       Delete;      // Backspace
     '_':      ChangeSign;  // Underscore
     ',',
     '.':      setPoint;
     '=',
     ^M,
     ^I:       Operate (_none);
   END (* CASE Key *);
 END;

PROCEDURE NoCalculator.Process (CONST Expr: string);
 VAR
   i: integer;
 BEGIN
   FOR i:= 1 TO Length (Expr) DO Process (Expr [i]);
 END;

PROCEDURE NoCalculator.Process (Key: keyty);
 BEGIN
   CASE Key OF
     Key_0..Key_9:  NextDigit (Ord (Key));  // chars [1]
     Key_Slash:     Operate (_div);
     Key_Asterisk:  Operate (_mul);
     Key_Minus:     Operate (_sub);
     Key_Plus:      Operate (_add);
     Key_Delete:    Clear;
     Key_Backspace: Delete;
     Key_Next:      ChangeSign;
     Key_Comma,
     Key_Period:    setPoint;
     Key_Return,
     Key_Equal,
     Key_Tab,
     Key_BackTab:   Operate (_none);
   END (* CASE Key *);
 END;

PROCEDURE NoCalculator.KeyAction (VAR KeyInfo: keyeventinfoty);
 BEGIN
   WITH KeyInfo DO BEGIN
     Process (Key);
     IF (Key IN [Key_0..Key_9, Key_Slash, Key_Asterisk, Key_Minus, Key_Plus, Key_Comma, Key_Period, Key_Equal]) OR
        (Key = Key_Delete) OR (Key = Key_Backspace) OR (Key = Key_Next) OR (Key = Key_Return) OR
        (Key = Key_Home) OR (Key = Key_End) {ignore these!}
     THEN BEGIN
       Key:= Key_none; chars:= '';
     END (* IF Key IN [...] *);
   END (* WITH aInfo *);
 END;

///////////////////////////////////////////////////////

FUNCTION IntCalculator.getValue: integer;
 BEGIN
   getValue:= Value;
 END;

PROCEDURE IntCalculator.setValue (NewValue: integer);
 begin
   currValue:= NewValue;
   setText (IntToStr (currValue));
 END;

PROCEDURE IntCalculator.Clear;
 BEGIN
   Value:= 0; Stack [0]:= 0; Stack [1]:= 0;
   prevOp [0]:= _none; prevOp [1]:= _none;
 END;

PROCEDURE IntCalculator.Delete;
 BEGIN
   Value:= Value DIV 10;
 END;

PROCEDURE IntCalculator.ChangeSign;
 BEGIN
   Value:= -Value;
 END;

PROCEDURE IntCalculator.nextDigit (Digit: integer);
 BEGIN
   IF currOp <> _none THEN BEGIN
     currOp:= _none; Value:= Digit;
   END (* IF currOp <> _none *)
   ELSE Value:= 10* Value+ Digit;
 END;
PROCEDURE IntCalculator.nextDigit (inChar: char);
 BEGIN
   nextDigit (ord (inChar)- ord ('0'));
 END;

//////////////////////////

PROCEDURE IntCalculator.Operate (OpSign: Operation);
{ Bearbeitung:
     wenn neue Operation
     +-:
         wenn letzte Operation
         +-:
             letzte Operation ausfuehren, neue Operation merken
         */:
             letzte Operation ausfuehren
             vorherige Operation pruefen (muss +- sein)
                 falls vorhanden, ausfuehren und verrechnen
             neue Operation merken
     */:
         wenn letzte Operation
         +-:
             letzte Operation ausfuehren, Ergebnis speichern
             neue Operation merken
         */:
             letzte Operation ausfuehren, neue Operation merken

  Beispiel:  1 + 2 * 3 + 4 * 5 + 6 = 33
  Eingabe Operation  Anzeige Stack0 Stack1 Operation0 Operation1
     1       .          1      0      0        .          .
     +       +          1      1      0        +          .
     2       .          2      1      0        +          .
     *       *          2      2      1        *          +
     3       .          3      2      1        *          +
     +       +          6      1      0        +          .   sofort weiter:
             +          7      7      0        +          .
     4       .          4      7      0        +          .
     *       *          4      4      7        *          +
     5       .          5      4      7        *          +
     +       +         20      7      0        +          .   sofort weiter:
             +         27     27      0        +          .
     6       .          6     27      0        +          .
     =       .         33      0      0        .          .
}
 PROCEDURE newOp (Op: Operation);
  BEGIN
    Stack [0]:= Value; prevOp [1]:= _none;
    prevOp [0]:= OpSign;
    currOp:= prevOp [0];
  END (* newOp *);

 BEGIN
   IF currOp = _none THEN
     CASE OpSign OF
       _none,
       _done,
       _add,
       _sub:
         CASE prevOp [0] OF
           _none:
             BEGIN
               newOp (_none);
             END (* _mul *);
           _add:
             BEGIN
               Value:= Stack [0]+ Value; newOp (_none);
             END (* _add *);
           _sub:
             BEGIN
               Value:= Stack [0]- Value; newOp (_none);
             END (* _sub *);
           _mul:
             BEGIN
               Value:= Stack [0]* Value+ Stack [1];
               Stack [1]:= 0; newOp (_none);
             END (* _mul *);
           _div:
             BEGIN
               Value:= Stack [0] DIV Value+ Stack [1];
               Stack [1]:= 0; newOp (_none);
             END (* _div *);
         END (* CASE Operation *);
       _mul,
       _div:
         CASE prevOp [0] OF
           _none:
             BEGIN
               newOp (_none);
             END (* _mul *);
           _add,
           _sub:
             BEGIN
               Stack [1]:= Stack [0]; newOp (prevOp [0]);
             END (* _add *);
           _mul:
             BEGIN
               Value:= Stack [0]* Value; newOp (prevOp [1]);
             END (* _mul *);
           _div:
             BEGIN
               Value:= Stack [0] DIV Value; newOp (prevOp [1]);
             END (* _div *);
         END (* CASE Operation *);
     END (* CASE Operation *);
 END (* Operate *);

///////////////////////////////////////////////////////

CONSTRUCTOR FloatCalculator.Create;
 BEGIN
   Clear;
 END;

FUNCTION FloatCalculator.getValue: double;
 BEGIN
   getValue:= Value;
 END;

PROCEDURE FloatCalculator.setValue (NewValue: double);
 BEGIN
   currValue:= NewValue;
   IF pointpos > 0
     THEN setText (FloatToStrF (currValue, ffFixed, 0, pointpos))
     ELSE IF pointpos < 0
       THEN setText (FloatToStrF (currValue, ffFixed, 0, 0))
       ELSE setText (FloatToStrF (currValue, ffFixed, 0, 0)+ DefaultFormatSettings.DecimalSeparator)
 END;

PROCEDURE FloatCalculator.Clear;
 BEGIN
   pointpos:= -1; Value:= 0.0;
   decSign [0]:= 0; Stack [0]:= 0.0;
   decSign [1]:= 0; Stack [1]:= 0.0;
   prevOp [0]:= _none; prevOp [1]:= _none;
 END;

PROCEDURE FloatCalculator.Delete;
 VAR
   rounding: TFPURoundingMode;
 BEGIN
//   IF pointpos > 0 THEN Dec (pointpos);

   IF pointpos > 0 THEN BEGIN
      Dec (pointpos);
      rounding:= setRoundMode (rmTruncate);
      Value:= RoundTo (Value, -pointpos);
      setRoundMode (rounding);
   END
   ELSE IF pointpos = 0 THEN pointpos:= -1
     ELSE Value:= Int (Value* 0.1);
 END;

PROCEDURE FloatCalculator.ChangeSign;
 BEGIN
   Value:= -Value;
 END;

PROCEDURE FloatCalculator.setPoint;
 BEGIN
   IF pointpos < 0 THEN pointpos:= 0;
 END;

PROCEDURE FloatCalculator.nextDigit (Digit: integer);
 BEGIN
   IF currOp <> _none THEN BEGIN
     currOp:= _none; pointpos:= -1;
     Value:= Digit;
   END (* IF Stack [0] <> _none *)
   ELSE
     IF pointpos >= 0 THEN BEGIN
       Inc (pointpos);
       Value:= RoundTo (Value+ (Digit* 10.0**-pointpos), -pointpos);
     END (* IF decimals < 1.0 *)
     ELSE BEGIN
       pointpos:= -1;
       Value:= (10.0* Value)+ Digit;
     END;
 END;
PROCEDURE FloatCalculator.nextDigit (inChar: char);
 BEGIN
   nextDigit (ord (inChar)- ord ('0'));
 END;

//////////////////////////

PROCEDURE FloatCalculator.Operate (OpSign: Operation);

 PROCEDURE newOp (Op: Operation);
  BEGIN
    Stack [0]:= Value;
    IF pointpos > 0 THEN decSign [0]:= pointpos
    ELSE BEGIN
      decSign [0]:= 0; pointpos:= -1;
    END;
    prevOp [1]:= Op;
    prevOp [0]:= OpSign;
    currOp:= prevOp [0];
  END (* newOp *);

 BEGIN
   IF currOp = _none THEN
     CASE OpSign OF
       _none,
       _done,
       _add,
       _sub:
         CASE prevOp [0] OF
           _none:
             BEGIN
               newOp (_none);
             END (* _mul *);
           _add:
             BEGIN
               IF pointpos < decSign [0] THEN pointpos:= decSign [0];
               Value:= Stack [0]+ Value;
               newOp (_none);
             END (* _add *);
           _sub:
             BEGIN
               IF pointpos < decSign [0] THEN pointpos:= decSign [0];
               Value:= Stack [0]- Value;
               newOp (_none);
             END (* _sub *);
           _mul:
             BEGIN
               IF pointpos > 0
                 THEN pointpos:= pointpos+ decSign [0]
                 ELSE pointpos:= decSign [0];
               IF pointpos < decSign [1] THEN pointpos:= decSign [1];
               IF prevOp [1] = _add
                 THEN Value:= Stack [0]* Value+ Stack [1]
                 ELSE Value:= Stack [0]* Value- Stack [1];

               Stack [1]:= 0; decSign [1]:= 0;
               newOp (_none);
             END (* _mul *);
           _div:
             BEGIN
               pointpos:= ceil (log10 (Value));
               pointpos:= pointpos+ decSign [0];
               IF pointpos < decSign [1] THEN pointpos:= decSign [1];
               IF prevOp [1] = _add
                 THEN Value:= Stack [0]/ Value+ Stack [1]
                 ELSE Value:= Stack [0]/ Value- Stack [1];

               Stack [1]:= 0; decSign [1]:= 0;
               newOp (_none);
             END (* _div *);
         END (* CASE Operation *);
       _mul,
       _div:
         CASE prevOp [0] OF
           _none:
             BEGIN
               newOp (_none);
             END (* _mul *);
           _add,
           _sub:
             BEGIN
               Stack [1]:= Stack [0];
               decSign [1]:= decSign [0];
               newOp (prevOp [0]);
             END (* _add *);
           _mul:
             BEGIN
               pointpos:= pointpos+ decSign [0];
               Value:= Stack [0]* Value;
               newOp (prevOp [1]);
             END (* _mul *);
           _div:
             BEGIN
               pointpos:= ceil (log10 (Value));
               pointpos:= pointpos+ decSign [0];
               Value:= Stack [0]/ Value;
               newOp (prevOp [1]);
             END (* _div *);
         END (* CASE Operation *);
     END (* CASE Operation *);
 END (* Operate *);

END.
